import gc
from functools import wraps

def garbage_collector_at_the_end(func):
    """
    Decorador para ejecutar el garbage collector después de que la función decorada termine.

    Parameters:
    func (callable): La función que será decorada.

    Returns:
    callable: La función decorada.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)

        gc.collect()
        # print(f"Garbage collector ejecutado después de la función: {func.__name__}")

        return result

    return wrapper



from  tiktoken import encoding_for_model
from enum import StrEnum
from tiktoken.model import MODEL_TO_ENCODING
import json

class TokeniserHelperExceptionErrors(StrEnum):
     UNSUPPORTED_MODEL = "Unsupported model {}."
     UNSUPPORTED_MODEL_CAPABILITIES = "Unsupported model capabilities {}."

class ModelMaximumRequestTokens:
    def __init__(self,input:int,output:int) -> None:
        self.input=input
        self.output=output
    def to_dict(self):
        return self.__dict__
class ModelCapabilities:
    def __init__(self,name:str,version:str,maximum_training_date:str,maximum_request_tokens:ModelMaximumRequestTokens) -> None:
        self.name=name
        self.version=version
        self.maximum_training_date=maximum_training_date
        if isinstance(maximum_request_tokens,ModelMaximumRequestTokens):
            self.maximum_request_tokens=maximum_request_tokens
        else:        
            self.maximum_request_tokens=ModelMaximumRequestTokens(**maximum_request_tokens)
    def to_dict(self):
        tmp = self.__dict__
        tmp["maximum_request_tokens"]=tmp["maximum_request_tokens"].to_dict()
        return tmp
class TokensWithModelCapabilitiesValidation:
    def __init__(self,num_tokens:int,validation_success:bool,remaining_tokens:int,maximum_request_tokens:ModelMaximumRequestTokens) -> None:
        self.num_tokens=num_tokens
        self.validation_success=validation_success
        self.remaining_tokens = remaining_tokens
        if isinstance(maximum_request_tokens,ModelMaximumRequestTokens):
            self.maximum_request_tokens=maximum_request_tokens
        else:        
            self.maximum_request_tokens=ModelMaximumRequestTokens(**maximum_request_tokens)
    def to_dict(self):
        tmp = self.__dict__
        tmp["maximum_request_tokens"]=tmp["maximum_request_tokens"].to_dict()
        return tmp
MODEL_CAPABILITIES = [
    ModelCapabilities(name="gpt-4o",version="2024-05-13",maximum_training_date="Oct 2023", maximum_request_tokens=ModelMaximumRequestTokens(input=128000,output=4096)).to_dict(),
    ModelCapabilities(name="gpt-4o",version="2024-08-06",maximum_training_date="Oct 2023",maximum_request_tokens=ModelMaximumRequestTokens(input=128000,output=16384)).to_dict(),
    ModelCapabilities(name="gpt-35-turbo",version="0125",maximum_training_date="Sep 2021",maximum_request_tokens=ModelMaximumRequestTokens(input=16385,output=4096)).to_dict(),
]

class TokeniserHelper:
    @classmethod
    def get_supported_models(cls):
        """
        Gets a dictionary of the models supported by their respective encodings.

        :return: A dictionary where the keys are the names of the models and the values ​​are their corresponding encodings.
        :rtype: dict[str, str]

        :example:
        ```python
        from tokeniser_helper import TokeniserHelper

        supported_models = TokeniserHelper.get_supported_models()
        for model, encoding in supported_models.items():
            print(f"tiktoken Model: {model}, Encoding scheme: {encoding}")
        # tiktoken Model: gpt-4o, Encoding scheme: o200k_base
        # tiktoken Model: gpt-35-turbo, Encoding scheme: cl100k_base
        ```
        ...
        """
        supported_models = {}
        for model, encoding in MODEL_TO_ENCODING.items():
            supported_models[model] = encoding
        return supported_models

    @classmethod
    def get_model_capabilities(cls, model: str = None, version: str = None) -> dict:
        """
        Retrieves the capabilities of a model based on its name and version.

        This function generates a dictionary with the capabilities of models loaded from `MODEL_CAPABILITIES`. 
        If a model name and version are specified, it returns the capabilities for that specific model and version.
        If no model name or version is specified, it returns a dictionary with all available model capabilities.

        Values taken from:
        https://learn.microsoft.com/en-us/azure/ai-services/openai/concepts/models
        https://platform.openai.com/docs/models/gpt-4o

        :param model: The name of the model to search for. (optional)
        :param version: The version of the model to search for. (optional)
        :return: A dictionary containing:
            - The capabilities of the specified model if `model` and `version` are provided.
            - A dictionary with all model capabilities if `model` and `version` are not provided.
        :raises ValueError: If the specified model is not supported or not found in `MODEL_CAPABILITIES`.

        :example:
        ```python
        from tokeniser_helper import TokeniserHelper
        import json

        model_capabilities = TokeniserHelper.get_model_capabilities()
        print(json.dumps(model_capabilities,indent=4))

        model_capabilities = TokeniserHelper.get_model_capabilities("gpt-35-turbo","0125")
        print(json.dumps(model_capabilities,indent=4))
        ```
        """
        model_capabilities_json = {}
        format_name = "{}-{}"
        for item in MODEL_CAPABILITIES:
            model_capabilities = ModelCapabilities(**item)
            model_capabilities_json[format_name.format(model_capabilities.name, model_capabilities.version)] = model_capabilities.to_dict()
        if any([model is not None, version is not None]):
            response = model_capabilities_json.get(format_name.format(model, version))
            if response is None:
                raise ValueError(TokeniserHelperExceptionErrors.UNSUPPORTED_MODEL_CAPABILITIES.format(format_name.format(model, version)))
            else:
                return response
        else:
            return model_capabilities_json

    @classmethod
    def calculate_tokens(cls,message:str|dict, model: str = "gpt-3.5-turbo") -> int:
        """
        Counts the number of tokens in a given input, which can be a text string or a JSON object, using the specified model.

        :param message: The input data to count tokens for, which can be either a text string (str) or a JSON object (dict).
        :param model: The name of the model for which to count tokens. The default is "gpt-3.5-turbo".
        :return: The number of tokens in the input data.

        :raises ValueError: If the specified model is not supported.

        :example:
        ```python
        from tokeniser_helper import TokeniserHelper

        example01 = "Hello world ... !!!"
        num_tokens = TokeniserHelper.calculate_tokens(example01,"gpt-4o")
        print(f"num_tokens for example01 with model=gpt-4o: {num_tokens}")

        example02 = {
            "name"              : "Hello world program",
            "cwd"               : "/usr/bin/example",
            "supported_types"   : "srt,int,char"
        }
        num_tokens = TokeniserHelper.calculate_tokens(example02,"gpt-4o")
        print(f"num_tokens for example02 with model=gpt-4o: {num_tokens}")
        # num_tokens for example01 with model=gpt-4o: 6
        # num_tokens for example02 with model=gpt-4o: 26
        ```
        """
        try:
            if isinstance(message,dict):
                message=json.dumps(message)
            encoding = encoding_for_model(model)
            tokens = encoding.encode(message)
            return len(tokens)
        except KeyError:
            raise ValueError(TokeniserHelperExceptionErrors.UNSUPPORTED_MODEL.format(model))

    @classmethod
    def calculate_tokens_with_model_capabilities(cls,message:str|dict, model: str = "gpt-3.5-turbo", version: str = "0125") -> TokensWithModelCapabilitiesValidation:
        """
        Calculates the number of tokens in a given input and checks if it complies with the model's capabilities for the specified version.

        This function calculates the number of tokens required for a given message, compares it with the token limits of the specified model and version, 
        and returns an object indicating whether the message complies with the model's token limits.

        :param message: The input data to count tokens for, which can be either a text string (str) or a JSON object (dict).
        :param model: The name of the model to use for token calculation. The default is "gpt-3.5-turbo".
        :param version: The version of the model to use for validating token limits. The default is "0125".
        :return: An instance of `TokensWithModelCapabilitiesValidation` containing:
            - `num_tokens`: The number of tokens calculated for the input.
            - `validation_success`: A boolean indicating whether the number of tokens is within the model's limits.
            - `remaining_tokens`: The number of tokens remaining within the model's limit (positive value means within limits).
            - `maximum_request_tokens`: The maximum number of tokens allowed by the model for input.

        :raises ValueError: If the specified model or version is not supported or found in `MODEL_CAPABILITIES`.

        :example:
        ```python
        from tokeniser_helper import TokeniserHelper
        import json

        message = "Test message ... !!!" * 10000
        model_capabilities = TokeniserHelper.calculate_tokens_with_model_capabilities(message,"gpt-4o","2024-08-06")
        print(json.dumps(model_capabilities.to_dict(),indent=4))
        ```
        """
            
        num_tokens = cls.calculate_tokens(message,model)
        model_capabilities = ModelCapabilities(**cls.get_model_capabilities(model,version))
        validation_success = False
        remaining_tokens = model_capabilities.maximum_request_tokens.input - num_tokens
        if remaining_tokens > 0:
            validation_success = True
        return TokensWithModelCapabilitiesValidation(
            num_tokens=num_tokens,
            validation_success = validation_success,
            remaining_tokens = remaining_tokens,
            maximum_request_tokens=model_capabilities.maximum_request_tokens
        )



plain_text_base64 = "UkVHSVNUUk8gTm8uLEVYUEVESUVOVEUsUkFESUNBQ0lPTixGRUNIQSxWSUdFTkNJQSxWYWxvcmVzX0EsVmFsb3Jlc19CDQpSU0lBQTE1TTAwMzgzLDE5ODExMjEsMjAyMTExMjc1NjksMjAyMS8wNi8zMCwyMDI0LzA2LzE5LDEsNA0KLDAsMCwsLDIsNA0K"


from io import BytesIO
from logyca_ai.utils.constants.ocr import OCREngine, OCREngineSettings
from logyca_ai.utils.helpers.garbage_collector_helper import garbage_collector_at_the_end
from logyca_ai.utils.schemes.output.conversations import ImageBase64
from PIL import Image # Pillow
import base64
import fitz  # PyMuPDF
import os
import pytesseract

@garbage_collector_at_the_end
def extract_text_from_pdf_file(filename_full_path:str|BytesIO,advanced_image_recognition:bool=False,ocr_engine_path:str=None,output_temp_dir:str=None)->str:
    """
    Extracts text from a PDF file.

    :param filename_full_path: Full path to the PDF file from which to extract text or file loaded in BytesIO RAM memory.
    :type filename_full_path: str|BytesIO
    :param advanced_image_recognition: Indicates whether to perform text recognition on images within the PDF.
                               If True, OCR techniques will be used to extract text from images.
    :type advanced_image_recognition: bool
    :param ocr_engine_path: Path to the OCR executable. If provided, this path will be used instead of the default.
    :type ocr_engine_path: str, optional
    :param output_temp_dir: Temporary directory for storing output files.
                            If not provided, a default tmp temporary directory in the application root folder will be used.
    :type output_temp_dir: str, optional

    :return: Extracted text from the PDF file.
    :rtype: str

    :raises FileNotFoundError: If the specified PDF file is not found.
    :raises ValueError: If the OCR path is invalid.

    :example:

    # Example usage
    ```python
    text = extract_text_from_pdf_file('/tmp/example.pdf', advanced_image_recognition=True, ocr_engine_path='/usr/local/bin/tesseract', output_temp_dir='/tmp/tesseract_output')
    ```

    """

    if advanced_image_recognition:
        if ocr_engine_path is None:
            pytesseract.pytesseract.tesseract_cmd=OCREngine.get_binary_path()
        else:
            pytesseract.pytesseract.tesseract_cmd=ocr_engine_path

    if(isinstance(filename_full_path,str)):
        doc = fitz.open(filename_full_path, filetype="pdf")
    else:
        doc = fitz.open(stream=filename_full_path, filetype="pdf")

    text = ""
    if output_temp_dir is None:
        output_temp_dir=os.path.abspath(os.path.join(os.path.dirname(__file__),OCREngineSettings.TMP_DIR))
    if not os.path.exists(output_temp_dir):
        os.makedirs(output_temp_dir)

    for page_num in range(len(doc)):
        page = doc.load_page(page_num)

        text += page.get_text()
        if advanced_image_recognition:
            image_list = page.get_images(full=True)
            for img_index, img in enumerate(image_list):
                xref = img[0]
                base_image = doc.extract_image(xref)
                image_bytes = base_image["image"]

                image_path = os.path.join(output_temp_dir, f"image_{page_num+1}_{img_index+1}.png")
                with open(image_path, "wb") as img_file:
                    img_file.write(image_bytes)

                image = Image.open(image_path)
                ocr_text = pytesseract.image_to_string(image)
                text += ocr_text

                os.remove(image_path)

    doc.close()
    return text

@garbage_collector_at_the_end
def extract_images_from_pdf_file(filename_full_path:str|BytesIO)->list:
    """
    Image extract from a PDF file.

    :param filename_full_path: Full path to the PDF file from which to extract text or file loaded in BytesIO RAM memory.
    :type filename_full_path: str|BytesIO

    :return: Images from the PDF file.
    :rtype: list[base64]

    :raises FileNotFoundError: If the specified PDF file is not found.

    :example:

    # Example usage
    ```python
    image_list = extract_images_from_pdf_file('/tmp/example.pdf')
    ```

    """
    if(isinstance(filename_full_path,str)):
        doc = fitz.open(filename_full_path, filetype="pdf")
    else:
        doc = fitz.open(stream=filename_full_path, filetype="pdf")
    images = []

    for page_num in range(len(doc)):
        page = doc.load_page(page_num)

        image_list = page.get_images(full=True)
        for img_index, img in enumerate(image_list):
            xref = img[0]
            base_image = doc.extract_image(xref)
            image_bytes = base_image["image"]
            image_format = base_image["ext"]
            image_base64 = base64.b64encode(image_bytes).decode("utf-8")
            images.append(ImageBase64(
                image_base64=image_base64,
                image_format=image_format
                ).to_dict()
            )

    doc.close()
    return images




from docx import Document # python-docx
from enum import StrEnum
from io import BytesIO
from logyca_ai.utils.constants.ocr import OCREngine, OCREngineSettings
from logyca_ai.utils.helpers.garbage_collector_helper import garbage_collector_at_the_end
from logyca_ai.utils.schemes.output.conversations import ImageBase64
from openpyxl import load_workbook
from PIL import Image  # Pillow
import base64
import json
import os
import pandas as pd
import pytesseract

class ExcelSheetStates(StrEnum):
    HIDDEN = "hidden"
    VERY_HIDDEN = "veryHidden"
    VISIBLE = "visible"

@garbage_collector_at_the_end
def extract_text_from_docx_file(filename_full_path:str|BytesIO,advanced_image_recognition:bool=False,ocr_engine_path:str=None,output_temp_dir:str=None):
    """
    Extracts text from a DOCX file.

    :param filename_full_path: Full path to the DOCX file from which to extract text or file loaded in BytesIO RAM memory.
    :type filename_full_path: str|BytesIO
    :param advanced_image_recognition: Indicates whether to perform text recognition on images within the DOCX.
                               If True, OCR techniques will be used to extract text from images.
    :type advanced_image_recognition: bool
    :param ocr_engine_path: Path to the OCR executable. If provided, this path will be used instead of the default.
    :type ocr_engine_path: str, optional
    :param output_temp_dir: Temporary directory for storing output files.
                            If not provided, a default tmp temporary directory in the application root folder will be used.
    :type output_temp_dir: str, optional

    :return: Extracted text from the DOCX file.
    :rtype: str

    :raises FileNotFoundError: If the specified DOCX file is not found.
    :raises ValueError: If the OCR path is invalid.
    """

    if advanced_image_recognition:
        if ocr_engine_path is None:
            pytesseract.pytesseract.tesseract_cmd=OCREngine.get_binary_path()
        else:
            pytesseract.pytesseract.tesseract_cmd=ocr_engine_path

    doc = Document(filename_full_path)

    text = ""
    if output_temp_dir is None:
        output_temp_dir=os.path.abspath(os.path.join(os.path.dirname(__file__),OCREngineSettings.TMP_DIR))
    if not os.path.exists(output_temp_dir):
        os.makedirs(output_temp_dir)

    for paragraph in doc.paragraphs:
        text += paragraph.text + "\n"

    if advanced_image_recognition:
        for rel in doc.part.rels.values():
            if "image" in rel.target_ref:
                image_part = rel.target_part
                image_bytes = image_part.image.blob

                image = Image.open(BytesIO(image_bytes))
                ocr_text = pytesseract.image_to_string(image)
                text += ocr_text

    return text

@garbage_collector_at_the_end
def extract_images_from_docx_file(filename_full_path:str|BytesIO)->list:
    """
    Extracts text from a DOCX file.

    :param filename_full_path: Full path to the DOCX file from which to extract text or file loaded in BytesIO RAM memory.
    :type filename_full_path: str|BytesIO

    :return: Extracted text from the DOCX file.
    :rtype: list

    :raises FileNotFoundError: If the specified DOCX file is not found.
    :raises ValueError: If the OCR path is invalid.
    """

    doc = Document(filename_full_path)
    images = []

    for rel in doc.part.rels.values():
        if "image" in rel.target_ref:
            image_part = rel.target_part
            image_bytes = image_part.image.blob

            img = Image.open(BytesIO(image_bytes))
            image_format = img.format if img.format else "PNG"
            image_format = str(image_format).lower()
            buffered = BytesIO()
            img.save(buffered, format=image_format)
            image_base64 = base64.b64encode(buffered.getvalue()).decode("utf-8")
            images.append(ImageBase64(
                image_base64=image_base64,
                image_format=image_format
                ).to_dict()
            )
            del buffered

    return images

@garbage_collector_at_the_end
def extract_text_from_excel_file(filename_full_path: str|BytesIO, advanced_image_recognition: bool = False, ocr_engine_path: str = None, output_temp_dir: str = None,format_output:str="json",only_sheet_visibility_enabled:bool=True):
    """
    Extracts text from an Excel file including all sheets and any embedded images.

    :param filename_full_path: Full path to the Excel file from which to extract text or file loaded in BytesIO RAM memory.
    :param advanced_image_recognition: Indicates whether to perform OCR on images within the Excel file.
    :param ocr_engine_path: Path to the OCR executable.
    :param output_temp_dir: Temporary directory for storing output files.
    :param format_output: Data array format for output, can be csv, json, list. The default is json.
    :param only_sheet_visibility_enabled: When looping through excel sheets, filter those with visible status if true, if false, do not loop through sheets with another status.
    :return: A text in JSON format representing the structure of a Microsoft Excel spreadsheet and has the following format.
                Sheet_name: It is the name of the primary key that represents an object, within this object, there are several subkeys.
                Sheet_name: It is the name of the primary key.
                content: Contains a list of lists (a two-dimensional array) that represents the content of the spreadsheet cells.
                image_text: Contains the text extracted from images within the Microsoft Excel document.
                Example of the content in the list of formats that I will send you.
                {
                    "sheet_name": {
                        "content": [["", "", "", ""],["", "", "", ""]],
                        "image_text": ""
                        },
                }
    :rtype: dict
    """
    # Set OCR engine path
    if advanced_image_recognition:
        if ocr_engine_path is None:
            pytesseract.pytesseract.tesseract_cmd = OCREngine.get_binary_path()
        else:
            pytesseract.pytesseract.tesseract_cmd = ocr_engine_path

    if output_temp_dir is None:
        output_temp_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), OCREngineSettings.TMP_DIR))
    if not os.path.exists(output_temp_dir):
        os.makedirs(output_temp_dir)

    result = {}
    sheets_description = {}

    ################################################
    # OCR - Extract data from images

    workbook = load_workbook(filename_full_path, data_only=True)
    for sheet_name in workbook.sheetnames:
        sheet = workbook[sheet_name]
        if only_sheet_visibility_enabled is False:
            make_extraction = True
        elif sheet.sheet_state == ExcelSheetStates.VISIBLE:
            make_extraction = True
        else:
            make_extraction = False

        if make_extraction is True:
            text = ""
            if advanced_image_recognition:
                for image in sheet._images:
                    img = Image.open(BytesIO(image._data()))
                    ocr_text = pytesseract.image_to_string(img)
                    text += ocr_text

                    text += "\n"
            result[sheet_name] = {
                "image_text": text
            }
            sheets_description[sheet_name] = {
                "visibility": sheet.sheet_state
            }
            del sheet
    del workbook

    ################################################
    # Read text data
    excel_data = pd.ExcelFile(filename_full_path)
    # print(f"memory get_memory_consumed_by_current_process:{get_memory_consumed_by_current_process()}")
    sheet_names = excel_data.sheet_names
    for sheet in sheet_names:

        if sheets_description.get(sheet,False):
            make_extraction = True
        else:
            make_extraction = False
        if make_extraction is True:
            df = pd.read_excel(filename_full_path, sheet_name=sheet, header=None)
            df = df.fillna("")
            df = df.dropna(how='all')

            if format_output.lower() == 'list':
                list_of_lists = df.values.tolist()
            elif format_output.lower() == 'csv':
                list_of_lists = df.to_csv(index=False, sep=";", lineterminator='\n')
            elif format_output.lower() == 'json':
                list_of_lists = df.to_json(orient='records')
            else:
                raise ValueError(f"Unsupported format format_output: {format_output}")

            # print(list_of_lists)
            # print(df)
            result[sheet]["content"]=list_of_lists
            del df
            del list_of_lists

    excel_data.close()
    del excel_data

    ################################################
    # Finish

    json_result = json.dumps(result,default=str)
    # print(json_result)
    return json_result

@garbage_collector_at_the_end
def extract_images_excel_file(filename_full_path: str|BytesIO,only_sheet_visibility_enabled:bool=True)->list:
    """
    Extracts text from an Excel file including all sheets and any embedded images.

    :param filename_full_path: Full path to the Excel file from which to extract text or file loaded in BytesIO RAM memory.
    :rtype: list
    :param only_sheet_visibility_enabled: When looping through excel sheets, filter those with visible status if true, if false, do not loop through sheets with another status.
    :rtype: bool
    """

    workbook = load_workbook(filename_full_path, data_only=True)
    images = []
    for sheet_name in workbook.sheetnames:
        sheet = workbook[sheet_name]
        
        if only_sheet_visibility_enabled is False:
            make_extraction = True
        elif sheet.sheet_state == ExcelSheetStates.VISIBLE:
            make_extraction = True
        else:
            make_extraction = False
        
        if make_extraction is True:
            for image in sheet._images:
                img = Image.open(BytesIO(image._data()))
                image_format = img.format if img.format else "PNG"
                image_format = str(image_format).lower()
                buffered = BytesIO()
                img.save(buffered, format=image_format)
                image_base64 = base64.b64encode(buffered.getvalue()).decode("utf-8")
                images.append(ImageBase64(
                    image_base64=image_base64,
                    image_format=image_format,
                    sheet_name=sheet_name
                    ).to_dict()
                )
                del buffered
        del sheet
    del workbook

    return images
